import pandas as pd

from learntools_dados_ufv.core import *

# q1
class Q1(CodingProblem):
    _var = 'series'
    _hint = 'Utilize a função `pd.Series` com uma lista de números aleatórios como argumento. A função `random.sample` pode ser útil...'

    _solution = ''
#     _solution = CS('''\
# # O segundo paramêtro da função range() não é inclusivo, por isso deve-se usar 1001 e não 1000.
# values = random.sample(range(-1000, 1001), 500)
# series = pd.Series(values)
# series
#     ''')

    def check(self, series):
        assert_isinstance(pd.Series, series, name=self._var)
        series_len = len(series)
        assert series_len == 500, ('Esperado que a variável `series` tenha uma série com 500 valores mas possui {}'.format(series_len))
        series_max = series.max()
        assert series_max <= 1000, ('Esperado que a variável `series` tenha uma série com um valor máximo de até **1000**, mas possui um valor máximo de **{}**'.format(series_max))
        series_min = series.min()
        assert series_min >= -1000, ('Esperado que a variável `series` tenha uma série com um valor mínimo de pelo menos **-1000**, mas possui um valor mínimo de **{}**'.format(series_min))

# q2
class Q2(CodingProblem):
    _vars = ['series', 'series_max']
    _hint = 'Utilize a função `.max()` da classe Series do Pandas'

    _solution = ''
#     _solution = CS('''\
# series_max = series.max()
# print(series_max)
#     ''')

    def check(self, series, series_max):
        assert_isinstance(int, series_max, name=self._var)
        correct_vaue = series.max()
        assert series_max == correct_vaue, ('A variável `series_max` (valor = `{}`) não possui o valor máximo da série `series`'.format(series_max))

# q3
class Q3(CodingProblem):
    _vars = ['series', 'series_min']
    _hint = 'Utilize a função `.min()` da classe Series do Pandas'

    _solution = ''
#     _solution = CS('''\
# series_min = series.min()
# print(series_min)
#     ''')

    def check(self, series, series_min):
        assert_isinstance(int, series_min, name=self._var)
        correct_vaue = series.min()
        assert series_min == correct_vaue, ('A variável `series_min` (valor = `{}`) não possui o valor mínimo da série `series`'.format(series_min))

# q4
class Q4(CodingProblem):
    _vars = ['series', 'series_positives']
    _hint = 'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.'

    _solution = ''
#     _solution = CS('''\
# series_positives = series[series > 0]
# print(series_positives)
#     ''')

    def check(self, series, series_positives):
        assert_isinstance(pd.Series, series_positives, name=self._var)
        assert_series_equals(series_positives, series[series > 0])

# q5
class Q5(CodingProblem):
    _vars = ['series', 'series_odds']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Um dado é ímpar quando não é divsível por 2 (resto da divisão != 0)',
        'Você também pode usar a função `pd.Series.apply` passando uma função com o retorno se um dado item da série deve ser incluso na filtragem ou não'
    ]

    _solution = ''
#     _solution = CS('''\
# series_odds = series[series % 2 == 1]
# # ou...
# series_odds = series[series % 2 != 0]
# # ou...
# def is_even(number):
#     return (number % 2) != 0

# series_odds = series[series.apply(is_even)]

# print(series_odds)
#     ''')

    def check(self, series, series_odds):
        assert_isinstance(pd.Series, series_odds, name=self._var)
        assert_series_equals(series_odds, series[series % 2 == 1])

# q6
class Q6(CodingProblem):
    _vars = ['series', 'series_range']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Você pode usar a sintaxe de acesso à listas de python lista[n:m] onde n é o número inicial e m o final',
        'O número final de um range python é não inclusivo!'
    ]

    _solution = CS('''\
series_range = series[50:101]

print(series_odds)
    ''')

    def check(self, series, series_range):
        assert_isinstance(pd.Series, series_range, name=self._var)
        assert_series_equals(series_range, series[50:101])

# q7
class Q7(CodingProblem):
    _vars = ['series', 'series_even_size']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Um dado é par quando é divsível por 2 (resto da divisão == 0)',
        'Você pode usar a mesma função python para saber o tamanho de uma lista, ou utilizar uma propridade `size` do `pd.Series`.'
    ]

    _solution = ''
#     _solution = CS('''\
# series_even_size = series[series % 2 == 0].size

# print(series_even_size)
#     ''')

    def check(self, series, series_even_size):
        assert_isinstance(int, series_even_size, name=self._var)
        assert series_even_size == series[series % 2 == 0].size

# q8
class Q8(CodingProblem):
    _vars = ['series', 'series_high_values_size']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Você pode usar a mesma função python para saber o tamanho de uma lista, ou utilizar uma propridade `size` do `pd.Series`.'
    ]

    _solution = ''
#     _solution = CS('''\
# series_high_values_size = series[series > 500].size

# print(series_high_values_size)
#     ''')

    def check(self, series, series_high_values_size):
        assert_isinstance(int, series_high_values_size, name=self._var)
        assert series_high_values_size == series[series > 500].size

# q9
class Q9(CodingProblem):
    _vars = ['series', 'series_subset_values']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`. Você pode usar múltiplas condições no filtro usando operadores booleanos.',
        'Talvez você precise utilizar parenteses para isolar os filtros...',
    ]

    _solution = ''
#     _solution = CS('''\
# series_subset_values = series[(series > 500) & (series < 700)]

# print(series_subset_values)
#     ''')

    def check(self, series, series_subset_values):
        assert_isinstance(pd.Series, series_subset_values, name=self._var)
        assert_series_equals(series_subset_values, series[(series > 500) & (series < 700)])

# q10
class Q10(CodingProblem):
    _vars = ['series', 'series_sqrt_values']
    _hint = 'Talvez a função `pd.Series.apply` seja útil... Ou talvez você possa fazer operações matemáticas diretamente na série...'

    _solution = ''
#     _solution = CS('''\
# series_sqrt_values = series ** 0.5

# print(series_sqrt_values)
#     ''')

    def check(self, series, series_sqrt_values):
        assert_isinstance(pd.Series, series_sqrt_values, name=self._var)
        assert_series_equals(series_sqrt_values, series ** 0.5)


# q11.a
class Q11A(CodingProblem):
    _var = 'grades'
    _hint = 'Utilize a função `pd.Series` com uma lista de números como argumento. A função `random.sample` pode ser útil...'

    _solution = ''
#     _solution = CS('''\
# # Você vai precisar criar sua própria estratégia para os valores `values`.
# grades = pd.Series(values)
#         ''')

    def check(self, grades):
        assert_isinstance(pd.Series, grades, name=self._var)

# q11.b
class Q11B(ThoughtExperiment):
    _hint = 'Que tal você olhar nos exercícios de 1 a 10? Pense como um pedagogo ou diretor da escola. Quais informações, mesmo que simples, você iria querer saber sobre seus alunos?'
    # _solution = ('Séries em Pandas possui várias funções úteis, para recuperar valores máximos, mínimos, média, mediana, etc. Essas informações são ricas em qualidade dependendo do contexto, mesmo que simples!')
    _solution = ''

# q11.c
class Q11C(ThoughtExperiment):
    _hint = 'Pense em quais respostas sobre o contexto de turmas de 5ª série apresentado uma série **não** consegue responder?'
    # _solution = ('Séries são úteis, mas são limitadas. Você não consegue extrair informação sobre turmas individuais, por exemplo. Essa questão é para debatermos, então não há bem uma resposta correta. Pense em mais limitações!')
    _solution = ''

# q11
Q11 = MultipartProblem(Q11A, Q11B, Q11C)

qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6,
    Q7,
    Q8,
    Q9,
    Q10,
    Q11,
], var_format='q{n}')

__all__ = list(qvars)



"""
learntools_dados_ufv.core exports all the content from the core sub-package that is typically
needed by an exercise module. For example, Problem subclasses to inherit from, 
assert functions to call from custom check() methods, the CodeSolution class (by convention,
aliased to CS for brevity) for wrapping _solution attributes, etc.

For most exercise modules, "from learntools_dados_ufv.core import *" should include all batteries needed.
"""
# This one isn't used within exercise modules, but we export it here so that exercise
# notebooks can begin with "from learntools_dados_ufv.core import binder"
from learntools_dados_ufv.core.globals_binder import binder

# All exercise modules conclude with a call to this helper function
from learntools_dados_ufv.core.utils import bind_exercises
# Problem and all its subclasses (e.g. EqualityCheckProblem, ThoughtExperiment)
from learntools_dados_ufv.core.problem import *
from learntools_dados_ufv.core.multiproblem import MultipartProblem
# This is used so frequently, it's worth giving it a short alias.
from learntools_dados_ufv.core.richtext import CodeSolution as CS
# Assert helper functions, often used in custom check methods on Problem subclasses
from learntools_dados_ufv.core.asserts import *

# In starter code for exercises where the user has to create a variable foo with
# some value, we start them off with `foo = __`. 
from learntools_dados_ufv.core.constants import PLACEHOLDER as ____

# A bit hacky, but we want to make sure '____' is imported when exercise modules import *
# from learntools_dados_ufv.core, and by default names beginning with an underscore aren't brought
# in by star imports.
__all__ = [name for name in dir() if not name.startswith('_')] + ['____']



import pandas as pd

from learntools_dados_ufv.core import *
from scipy.stats import percentileofscore

# q1
class Q1(CodingProblem):
    _var = 'population_weight'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_populacao.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, population_weight):
        assert_isinstance(pd.Series, population_weight, name='population_weight')
        assert_series_equals(population_weight, pd.read_csv('./pesos_populacao.csv', header=None, squeeze=True), name='population_weight')

# q2 - a
class Q2A(CodingProblem):
    _var = 'sample_1_weight'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_amostra1.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, sample_1_weight):
        assert_isinstance(pd.Series, sample_1_weight, name='sample_1_weight')
        assert_series_equals(sample_1_weight, pd.read_csv('./pesos_amostra1.csv', header=None, squeeze=True), name='sample_1_weight')

# q2 - b
class Q2B(EqualityCheckProblem):
    _var = 'is_sample_1_from_population'
    _expected = True
    _hints = [
        'Calcule a média da população e da amostra, como no roteiro',
        'Após calcular as médias, calcule o erro padrão.',
        'Veja a diferença entre a média da amostra e da população, em erros padrão',
        'Existe algum valor estatístico de referência que conseguimos comparar pra ter uma certa certeza que uma amostra pertence a população?',
        'Salve o resultado em uma variável booleana, como `is_sample_1_from_population = False` ou `is_sample_1_from_population = True`'
    ]
    _solution = ''

# q2
Q2 = MultipartProblem(Q2A, Q2B)

# q3 - a
class Q3A(CodingProblem):
    _var = 'sample_2_weight'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_amostra2.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, sample_2_weight):
        assert_isinstance(pd.Series, sample_2_weight, name='sample_2_weight')
        assert_series_equals(sample_2_weight, pd.read_csv('./pesos_amostra2.csv', header=None, squeeze=True), name='sample_2_weight')

# q3 - b
class Q3B(EqualityCheckProblem):
    _var = 'is_sample_2_from_population'
    _expected = False
    _hints = [
        'Calcule a média da população e da amostra, como no roteiro',
        'Após calcular as médias, calcule o erro padrão.',
        'Veja a diferença entre a média da amostra e da população, em erros padrão',
        'Existe algum valor estatístico de referência que conseguimos comparar pra ter uma certa certeza que uma amostra pertence a população?',
        'Salve o resultado em uma variável booleana, como `is_sample_2_from_population = False` ou `is_sample_2_from_population = True`'
    ]
    _solution = ''

# q3
Q3 = MultipartProblem(Q3A, Q3B)

# q4 - a
class Q4A(CodingProblem):
    _var = 'sample_3_weight'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_amostra3.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, sample_3_weight):
        assert_isinstance(pd.Series, sample_3_weight, name='sample_3_weight')
        assert_series_equals(sample_3_weight, pd.read_csv('./pesos_amostra3.csv', header=None, squeeze=True), name='sample_3_weight')

# q4 - b
class Q4B(EqualityCheckProblem):
    _var = 'is_sample_3_from_population'
    _expected = False
    _hints = [
        'Calcule a média da população e da amostra, como no roteiro',
        'Após calcular as médias, calcule o erro padrão.',
        'Veja a diferença entre a média da amostra e da população, em erros padrão',
        'Existe algum valor estatístico de referência que conseguimos comparar pra ter uma certa certeza que uma amostra pertence a população?',
        'Salve o resultado em uma variável booleana, como `is_sample_3_from_population = False` ou `is_sample_3_from_population = True`'
    ]
    _solution = ''

# q4
Q4 = MultipartProblem(Q4A, Q4B)

# q5 - a
class Q5A(CodingProblem):
    _var = 'sample_4_weight'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_amostra4.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, sample_4_weight):
        assert_isinstance(pd.Series, sample_4_weight, name='sample_4_weight')
        assert_series_equals(sample_4_weight, pd.read_csv('./pesos_amostra4.csv', header=None, squeeze=True), name='sample_4_weight')

# q5 - b
class Q5B(EqualityCheckProblem):
    _var = 'is_sample_4_from_population'
    _expected = False
    _hints = [
        'Calcule a média da população e da amostra, como no roteiro',
        'Após calcular as médias, calcule o erro padrão.',
        'Veja a diferença entre a média da amostra e da população, em erros padrão',
        'Existe algum valor estatístico de referência que conseguimos comparar pra ter uma certa certeza que uma amostra pertence a população?',
        'Salve o resultado em uma variável booleana, como `is_sample_4_from_population = False` ou `is_sample_4_from_population = True`'
    ]
    _solution = ''

# q5
Q5 = MultipartProblem(Q5A, Q5B)

# q6
class Q6(ThoughtExperiment):
    _hints = [
        'Veja os resultados dessas amostras se pertencem a uma mesma população em comum, nas atividades anteriores.',
    ]
    _solution = ''

qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6
], var_format='q{n}')

__all__ = list(qvars)



# This file defines ____ which we use in courses as a sentinel indicating a user hasn't
# attempted an exercise.

class PlaceholderValue(object):

    def __eq__(self, other):
        return isinstance(other, PlaceholderValue)

    def _repr_markdown_(self):
        """This returns the empty string. Some questions' starter code cell will look like:
            foo = ____
            q1.check()
            foo
        i.e. we want to echo the value of the variable at the end to make sure the user
        understands what kind of object they've created. While the variable still has 
        its placeholder value, we shouldn't echo anything.
        """
        return ''

PLACEHOLDER = PlaceholderValue()



import pandas as pd
from math import floor


from learntools_dados_ufv.core import *

# q1
class Q1(CodingProblem):
    _var = 'df'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `weatherHistory.csv`'

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_df_equals(df, pd.read_csv('./League_Games.csv'), name='df')

# q2
class Q2(CodingProblem):
    _vars= ['df', 'X_train', 'Y_train', 'X_test', 'Y_test']
    _hint = 'Utilize a função `train_test_split` do módulo `sklearn.model_selection`.'

    _solution = ''

    def check(self, df, X_train, Y_train, X_test, Y_test):
        assert_isinstance(pd.DataFrame, X_train, name='X_train')
        assert_isinstance(pd.DataFrame, X_test, name='X_test')
        assert_isinstance(pd.Series, Y_train, name='Y_train')
        assert_isinstance(pd.Series, Y_test, name='Y_test')

        expected_length_train = floor(0.7 * len(df))
        expected_length_test =  len(df) - expected_length_train

        assert_len(X_train, expected_length_train, name='X_train')
        assert_len(Y_train, expected_length_train, name='Y_train')
        assert_len(X_test, expected_length_test, name='X_test')
        assert_len(Y_test, expected_length_test, name='Y_test')

# q3
class Q3(ThoughtExperiment):
    _hints = [
        'Faça um plot de distribuição com a coluna de interesse (Y).',
        'Revisite as práticas anteriores para entender como fazer uma análise gráfica de um modelo desconhecido.',
    ]

# q4
class Q4(ThoughtExperiment):
    _hints = [
        'Existem inúmeros algoritmos de IA supervisionados. Visite alguns vistos nas aulas téoricas presentes no `scikit-learn`.',
        'Verifique as condições que os algoritmos assumem que os dados possuem antes de os adotar.',
        'Instancie o modelo e faça o treinamento passando apenas o conjunto de treinamento.',
    ]

# q5
class Q5(ThoughtExperiment):
    _hints = [
        'Visite o roteiro para ver como é calculado a acurácia, precisão e revogação.',
        'Pesquise sobre os métodos presentes no módulo `sklearn.stats` e `sklearn.metrics`.',
    ]

# q6
class Q6(ThoughtExperiment):
    _hints = [
        'Tente aplicar o que aprendeu sobre identificar quais as features mais importantes em um conjunto de dados.',
        'Experimente o treinamento do modelo passando outros argumentos intrínsecos ao algoritmo.',
        'Experimente o treinamento do modelo passando um conjunto de atributos diferente.',
        'Experimente o treinamento do modelo tratando antes os dados de alguma maneira.',
    ]

# q7
class Q7(ThoughtExperiment):
    _hints = [
        'Faça novamente o treinamento e validação calculando o tempo gasto.',
        'Pesquise sobre como realizar um benchmark de tempo em python usando o pacote `time`.',
    ]

# q8
class Q8(ThoughtExperiment):
    _hints = [
        'Busque nos pacotes do `scikit-learn` uma forma fácil de calcular a matriz de confusão.',
    ]


qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6,
    Q7,
    Q8
], var_format='q{n}')

__all__ = list(qvars)



import pandas as pd

from learntools_dados_ufv.core import *

# q1
class Q1(CodingProblem):
    _var = 'df'
    _hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `weatherHistory.csv`'

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_df_equals(df, pd.read_csv('./weatherHistory.csv'), name='df')

# q2
class Q2(ThoughtExperiment):
    _hints = [
        'Aplique o K-Means utilizando todas as features válidas, exceto `Precip Type`, e obviamente a `Target`',
        'Utilize a mesma quantidade de clusters que possui nos grupos reais da base.',
    ]

# q3
class Q3(ThoughtExperiment):
    _hints = [
        'Revisite o roteiro e faça gráficos parecidos com os que estão lá.',
        'Descreva a diferença entre eles e o que significa cada grupo...',
        'Talvez seja útil utilizar a função `classification_report` do módulo `sklearn.metrics`',
    ]

# q4
class Q4(ThoughtExperiment):
    _hints = [
        'Tente visualizar a correlação de Pearson e utilizar as features com maior correlação com a precipitação',
        'Aplique o processo de clusterização e plotagem de gráfico para cada feature relevante.',
    ]

# q5
class Q5(ThoughtExperiment):
    _hints = [
        'Compare os valores do teste da silhueta, quais os melhores? E os piores, por que são piores?',
    ]

# q6
class Q6(ThoughtExperiment):
    _hints = [
        'Pesquise a resposta e verifique se já existe alguma função pronta nas bibliotecas utilizadas pra facilitar sua vida.',
    ]

# q7
class Q7(ThoughtExperiment):
    _hints = [
        'Pesquise a resposta e verifique se já existe alguma função pronta nas bibliotecas utilizadas pra facilitar sua vida.',
        'É esperado que seja analisado algumas facetas como tempo de execução, performance, tipo de agrupamento, premissas, etc.',
    ]


qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6,
    Q7,
], var_format='q{n}')

__all__ = list(qvars)



from urllib import response
import pandas as pd

from mlxtend.preprocessing import TransactionEncoder
from mlxtend.frequent_patterns import apriori
from mlxtend.frequent_patterns import association_rules

from learntools_dados_ufv.core import *

root_df = pd.read_excel('./OnlineRetail.xlsx')

def run_df_processing(steps=0):
    answer = root_df.copy()
    if (steps == 0): 
        return answer
    
    answer['Description'] = answer['Description'].str.strip()
    if (steps == 1): 
        return answer
    
    answer = answer.dropna(axis=0, subset=['InvoiceNo'])
    if (steps == 2): 
        return answer

    answer['InvoiceNo'] = answer['InvoiceNo'].astype('str')
    if (steps == 3): 
        return answer

    answer = answer[answer['Country'] == 'Germany']
    if (steps == 4): 
        return answer
    
    answer = answer.groupby(['InvoiceNo', 'Description'])
    if (steps == 5): 
        return answer
    
    answer = answer['Quantity'].sum()
    if (steps == 6): 
        return answer
    
    answer = answer.unstack().reset_index().fillna(0).set_index('InvoiceNo')
    if (steps == 7): 
        return answer

    def encode_units(x):
        if x <= 0:
            return 0
        if x >= 1:
            return 1

    answer = answer.applymap(encode_units)
    if (steps == 8): 
        return answer

    answer = answer.drop('POSTAGE', axis=1)
    if (steps == 9): 
        return answer

    answer = apriori(answer, min_support=0.04, use_colnames=True)
    if (steps == 10): 
        return answer
    
    rules = association_rules(answer, min_threshold=0.5)
    if (steps == 11): 
        return rules
    
    if (steps == 12): 
        return rules[(rules['support'] >= 0.015) & (rules['confidence'] >= 0.75)]
    
    if (steps == 13): 
        return rules[(rules['lift'] >= 4) & (rules['confidence'] >= 0.8)]

# q1.a
class Q1A(CodingProblem):
    _var = 'df'
    _hints = [
        'Pesquise sobre a função `strip` do tipo str em Python.',
        'Você pode usar as funções nativas do tipo str em uma coluna Pandas usando a sintaxe `dataframe["Coluna"].str.funcao()`.',
        'Altere uma coluna usando a sintaxe `df.Coluna = df.Coluna.algumacoisa()`.',
    ]

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')

        answer = run_df_processing(1)
        assert_df_equals(df, answer, name='df')

# q1.b
class Q1B(CodingProblem):
    _var = 'df'
    _hints = [
        'Pesquise sobre a função `pd.DataFrame.dropna` do tipo DataFrame Pandas.',
        'Pesquise sobre os parâmetros `subset` e `axis` da função `dropna`.',
        'Altere um dataframe usando a sintaxe `df = df.algumacoisa()`.',
    ]

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')

        answer = run_df_processing(2)
        assert_df_equals(df, answer, name='df')

# q1.c
class Q1C(CodingProblem):
    _var = 'df'
    _hints = [
        'Pesquise sobre a função `astype` de uma coluna de DataFrame Pandas.',
        'Converta para o tipo `str`',
    ]

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')

        answer = run_df_processing(3)
        assert_df_equals(df, answer, name='df')


# q1
Q1 = MultipartProblem(Q1A, Q1B, Q1C)

# q2.a
class Q2A(CodingProblem):
    _var = 'df_germany'
    _hints = [
        'Faça um filtro simples com `dataframe[filtro]`.',
        'Altere um dataframe usando a sintaxe `df = df[filtro]`.',
    ]

    _solution = ''

    def check(self, df_germany):
        assert_isinstance(pd.DataFrame, df_germany, name='df_germany')

        answer = run_df_processing(4)
        assert_df_equals(df_germany, answer, name='df_germany')

# q2.b
class Q2B(CodingProblem):
    _var = 'df_germany_grouped'
    _hints = [
        'Utilize a função `groupby` do Pandas.',
        'Você pode passar uma lista para a função `groupby`.',
        'A ordem da lista importa!',
    ]

    _solution = ''

    def check(self, df_germany_grouped):
        answer = run_df_processing(5)

        v = df_germany_grouped.apply(lambda x: x.equals(answer.get_group(x.name)) if x.name in answer.groups else False)
        failing_rows = len(v[v == False])

        assert failing_rows == 0, "O grupo fornecido possui pelo menos {} registros incorretos.".format(failing_rows)

# q2.c
class Q2C(CodingProblem):
    _var = 'df_germany_grouped_sum'
    _hints = [
        'Utilize a função `sum` de uma coluna agrupada do Pandas.',
        'Utilize a sintaxe `group[coluna].funcao()`',
    ]

    _solution = ''

    def check(self, df_germany_grouped_sum):
        answer = run_df_processing(6)
        assert_equal(df_germany_grouped_sum, answer, name='df_germany_grouped_sum')

# q2.d
class Q2D(CodingProblem):
    _var = 'df_germany_basket'
    _hints = [
        'Pesquise sobre a função `unstack` do Pandas.',
        'Pesquise sobre a função `reset_index` do Pandas.',
        'Pesquise sobre a função `fillna` do Pandas.',
        'Pesquise sobre a função `set_index` do Pandas.',
        'Utilize as dicas anteriores em sequência.',
    ]

    _solution = ''

    def check(self, df_germany_basket):
        assert_isinstance(pd.DataFrame, df_germany_basket, name='df_germany_basket')

        answer = run_df_processing(7)
        assert_df_equals(df_germany_basket, answer, name='df_germany_basket')

# q2.e
class Q2E(CodingProblem):
    _var = 'df_germany_basket'
    _hints = [
        'Crie uma função auxiliar que recebe um valor numérico e retorna 0 ou 1 com a condição esperada.',
        'Pesquise sobre a função `applymap` do Pandas.',
    ]

    _solution = ''

    def check(self, df_germany_basket):
        assert_isinstance(pd.DataFrame, df_germany_basket, name='df_germany_basket')

        answer = run_df_processing(8)
        assert_df_equals(df_germany_basket, answer, name='df_germany_basket')

# q2.f
class Q2F(CodingProblem):
    _var = 'df_germany_basket'
    _hints = [
        'Pesquise sobre a função `dataframe.drop` do Pandas.',
        'Você quer dropar uma **coluna**. Então, lembre-se de passar o parâmetro `axis` corretamente.',
    ]

    _solution = ''

    def check(self, df_germany_basket):
        assert_isinstance(pd.DataFrame, df_germany_basket, name='df_germany_basket')

        answer = run_df_processing(9)
        assert_df_equals(df_germany_basket, answer, name='df_germany_basket')

# q2
Q2 = MultipartProblem(Q2A, Q2B, Q2C, Q2D, Q2E, Q2F)

# q3.a
class Q3A(CodingProblem):
    _var = 'germany_frequent_itemsets'
    _hints = [
        'Utilize a função `apriori` como mostrado no roteiro.',
        'Você deve passar os paramêtros na descrição do problema.',
        'Você adicionou os parâmetros `use_colnames` e `min_support`?',
    ]

    _solution = ''

    def check(self, germany_frequent_itemsets):
        answer = run_df_processing(10)
        assert_equal(germany_frequent_itemsets, answer, name='germany_frequent_itemsets')

# q3.b
class Q3B(CodingProblem):
    _var = 'germany_rules'
    _hints = [
        'Utilize a função `association_rules` como mostrado no roteiro.',
        'Você deve passar os paramêtros na descrição do problema.',
        'Você adicionou o parâmetro `min_threshold`?',
    ]

    _solution = ''

    def check(self, germany_rules):
        answer = run_df_processing(11)
        assert_equal(germany_rules, answer, name='germany_rules')

# q3
Q3 = MultipartProblem(Q3A, Q3B)

# q4.a
class Q4A(CodingProblem):
    _var = 'germany_rules_a'
    _hints = [
        'Utilize um filtro, como mostrado no roteiro.',
    ]

    _solution = ''

    def check(self, germany_rules_a):
        assert_isinstance(pd.DataFrame, germany_rules_a, name='germany_rules_a')

        answer = run_df_processing(12)
        assert_df_equals(germany_rules_a, answer, name='germany_rules_a')

# q4.b
class Q4B(CodingProblem):
    _var = 'germany_rules_b'
    _hints = [
        'Utilize um filtro, como mostrado no roteiro.',
    ]

    _solution = ''

    def check(self, germany_rules_b):
        assert_isinstance(pd.DataFrame, germany_rules_b, name='germany_rules_b')

        answer = run_df_processing(13)
        assert_df_equals(germany_rules_b, answer, name='germany_rules_b')

# q4.c
class QBC(ThoughtExperiment):
    _hint = 'Cite 2 ou 3 pares/trios de produtos interpretando os resultados obtidos em `germany_rules`.'


# q4
Q4 = MultipartProblem(Q4A, Q4B)


qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
], var_format='q{n}')

__all__ = list(qvars)



import pandas as pd

from learntools_dados_ufv.core import *
from scipy.stats import percentileofscore

# q1
class Q1(CodingProblem):
    _vars = ['men_height_series', 'women_height_series']
    _hint = 'Utilize a função `pd.Series` com o caminho dos arquivos `altura_homens.csv` e `altura_mulheres.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, men_height_series, women_height_series):
        assert_isinstance(pd.Series, men_height_series, name='men_height_series')
        assert_isinstance(pd.Series, series_mulheres, name='women_height_series')

        series_homens = pd.read_csv("./altura_homens.csv",header= None)[0]
        series_mulheres = pd.read_csv("./altura_mulheres.csv",header= None)[0]
        assert_series_equals(men_height_series,series_homens , name='men_height_series')
        assert_series_equals(women_height_series, series_mulheres, name='women_height_series')

# q2
class Q2(CodingProblem):
    _vars = ['men_height_series', 'women_height_series', 'men_min_height', 'men_max_height', 'women_min_height', 'women_max_height']
    _hint = 'Utilize a função `.max()` e `.min()` da classe Series do Pandas'
    _solution = ''

    def check(self, 
            men_height_series,
            women_height_series,
            men_min_height,
            men_max_height,
            women_min_height,
            women_max_height
        ):
        assert_isinstance(float, men_min_height, name='men_min_height')
        assert_isinstance(float, men_max_height, name='men_max_height')
        assert_isinstance(float, women_min_height, name='women_min_height')
        assert_isinstance(float, women_max_height, name='women_max_height')
        assert men_min_height == men_height_series.min(), ('A variável `men_min_height` (valor = `{}`) não possui o valor mínimo da Series `men_height_series`'.format(men_min_height))
        assert men_max_height == men_height_series.max(), ('A variável `men_max_height` (valor = `{}`) não possui o valor máximo da Series `men_height_series`'.format(men_max_height))
        assert women_min_height == women_height_series.min(), ('A variável `women_min_height` (valor = `{}`) não possui o valor mínimo da Series `women_height_series`'.format(women_min_height))
        assert women_max_height == women_height_series.max(), ('A variável `women_max_height` (valor = `{}`) não possui o valor máximo da Series `women_height_series`'.format(women_max_height))

# q3
class Q3(CodingProblem):
    _vars = ['men_height_series', 'women_height_series', 'men_mean_height', 'women_mean_height']
    _hint = 'Utilize a função `.mean()` da classe Series do Pandas'

    _solution = ''

    def check(self,
            men_height_series,
            women_height_series,
            men_mean_height,
            women_mean_height
        ):
        assert_isinstance(float, men_mean_height, name='men_mean_height')
        assert_isinstance(float, women_mean_height, name='women_mean_height')
        assert men_mean_height == men_height_series.mean(), ('A variável `men_mean_height` (valor = `{}`) não possui o valor da média da Series `men_height_series`'.format(men_mean_height))
        assert women_mean_height == women_height_series.mean(), ('A variável `women_mean_height` (valor = `{}`) não possui o valor da média da Series `women_height_series`'.format(women_mean_height))

# q4
class Q4(CodingProblem):
    _vars = ['men_height_series', 'women_height_series', 'men_median_height', 'women_median_height']
    _hint = 'Utilize a função `.median()` da classe Series do Pandas'

    _solution = ''

    def check(self,
            men_height_series,
            women_height_series,
            men_median_height,
            women_median_height
        ):
        assert_isinstance(float, men_median_height, name='men_median_height')
        assert_isinstance(float, women_median_height, name='women_median_height')
        assert men_median_height == men_height_series.median(), ('A variável `men_median_height` (valor = `{}`) não possui o valor da mediana da Series `men_height_series`'.format(men_median_height))
        assert women_median_height == women_height_series.median(), ('A variável `women_median_height` (valor = `{}`) não possui o valor da mediana da Series `women_height_series`'.format(women_median_height))

# q5
class Q5(CodingProblem):
    _vars = ['men_height_series', 'women_height_series', 'men_std_height', 'women_std_height']
    _hint = 'Utilize a função `.std()` da classe Series do Pandas'

    _solution = ''

    def check(self,
            men_height_series,
            women_height_series,
            men_std_height,
            women_std_height
        ):
        assert_isinstance(float, men_std_height, name='men_std_height')
        assert_isinstance(float, women_std_height, name='women_std_height')
        assert men_std_height == men_height_series.std(), ('A variável `men_std_height` (valor = `{}`) não possui o valor do desvio padrão da Series `men_height_series`'.format(men_std_height))
        assert women_std_height == women_height_series.std(), ('A variável `women_std_height` (valor = `{}`) não possui o valor do desvio padrão da Series `women_height_series`'.format(women_std_height))

# q6
class Q6(CodingProblem):
    _vars = ['men_height_series', 'men_shorter_160cm_perc']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Divida o total filtrado pelo total da base para ter a proporção',
        'Não esqueça de multiplicar o resultado da divisão por 100...'
    ]

    _solution = ''

    def check(self,
            men_height_series,
            men_shorter_160cm_perc
        ):
        assert_isinstance(float, men_shorter_160cm_perc, name='men_shorter_160cm_perc')
        assert men_shorter_160cm_perc <= 100 and men_shorter_160cm_perc >= 0, ('A variável `men_shorter_160cm_perc` deve possuir um valor entre 0 e 100.')
        men_shorter_160cm = men_height_series[men_height_series < 160]
        result = 100 * (len(men_shorter_160cm) / len(men_height_series))
        assert men_shorter_160cm_perc == result, ('A variável `men_shorter_160cm_perc` (valor = `{}`) não possui o valor do percentual de homens com altura menor que 160cm da Series `men_height_series`'.format(men_shorter_160cm_perc))

# q7
class Q7(CodingProblem):
    _vars = ['women_height_series', 'women_taller_180cm_perc']
    _hints = [
        'Faça o filtro de acesso à serie usando a sintaxe `series[filtro]`.',
        'Divida o total filtrado pelo total da base para ter a proporção',
        'Não esqueça de multiplicar o resultado da divisão por 100...'
    ]

    _solution = ''

    def check(self,
            women_height_series,
            women_taller_180cm_perc
        ):
        assert_isinstance(float, women_taller_180cm_perc, name='women_taller_180cm_perc')
        assert women_taller_180cm_perc <= 100 and women_taller_180cm_perc >= 0, ('A variável `women_taller_180cm_perc` deve possuir um valor entre 0 e 100.')
        women_taller_180cm = women_height_series[women_height_series > 180]
        result = 100 * (len(women_taller_180cm) / len(women_height_series))
        assert women_taller_180cm_perc == result, ('A variável `women_taller_180cm_perc` (valor = `{}`) não possui o valor do percentual de homens com altura menor que 160cm da Series `women_height_series`'.format(women_taller_180cm_perc))

# q8
class Q8(CodingProblem):
    _vars = ['men_height_series', 'men_185cm_percentile']
    _hints = [
        'Pesquise sobre a função `percentileofscore` do pacote SciPy',
        'Realmente use a função `percentileofscore` do pacote SciPy'
    ]

    _solution = ''

    def check(self,
            men_height_series,
            men_185cm_percentile
        ):
        assert_isinstance(float, men_185cm_percentile, name='men_185cm_percentile')
        result = percentileofscore(men_height_series, 185)
        assert men_185cm_percentile == result, ('A variável `men_185cm_percentile` (valor = `{}`) não possui o valor do percentil de um homem com altura de 185cm'.format(men_185cm_percentile))

# q9
class Q9(CodingProblem):
    _vars = ['women_height_series', 'women_150cm_percentile']
    _hints = [
        'Pesquise sobre a função `percentileofscore` do pacote SciPy',
        'Realmente use a função `percentileofscore` do pacote SciPy'
    ]

    _solution = ''

    def check(self,
            women_height_series,
            women_150cm_percentile
        ):
        assert_isinstance(float, women_150cm_percentile, name='women_150cm_percentile')
        result = percentileofscore(women_height_series, 150)
        assert women_150cm_percentile == result, ('A variável `women_150cm_percentile` (valor = `{}`) não possui o valor do percentil de um homem com altura de 185cm'.format(women_150cm_percentile))

# q10
class Q10(CodingProblem):
    _vars = ['men_height_series', 'men_top3_height', 'men_in_top3_height']
    _hints = [
        'Pesquise sobre a função `value_counts` da classe Series do Pandas',
        'Pesquise sobre a função `head` da classe Series do Pandas',
        'Pesquise sobre a função `iloc` da classe Series do Pandas',
        'Some o total das pessoas nas 3 alturas!',
    ]

    _solution = ''

    def check(self,
            men_height_series,
            men_top3_height,
            men_in_top3_height
        ):
        assert_isinstance(pd.Series, men_top3_height, name='men_top3_height')
        assert_isinstance(int, men_in_top3_height, name='men_in_top3_height')
        vc = men_height_series.value_counts()
        for i, _row in men_top3_height.iteritems():
            assert i in vc, ('O valor {} não é uma das top 3 alturas dos homens'.format(i))

     
     
# q11
class Q11(CodingProblem):
    _vars = ['women_height_series', 'women_185cm_std_distance', 'women_145cm_std_distance']
    _hints = [
        'Pesquise sobre a função `std` da classe Series do Pandas para extrair o desvio padrão',
        'Pesquise sobre a função `mean` da classe Series do Pandas para extrair a média',
    ]

    _solution = ''

    def check(self,
            women_height_series,
            women_185cm_std_distance,
            women_145cm_std_distance
        ):
        assert_isinstance(float, women_185cm_std_distance, name='women_185cm_std_distance')
        assert_isinstance(float, women_145cm_std_distance, name='women_145cm_std_distance')
        
        mean = women_height_series.mean()
        std = women_height_series.std()

        d1 = abs(185 - mean) / std
        d2 = abs(145 - mean) / std

        assert women_185cm_std_distance == d1, ('O valor {} não é o resultado correto da quantidade de desvios padrão de distância da média'.format(women_185cm_std_distance))
        assert women_145cm_std_distance == d2, ('O valor {} não é o resultado correto da quantidade de desvios padrão de distância da média'.format(women_145cm_std_distance))

# q12
class Q12(ThoughtExperiment):
    _hint = 'Tente olhar para alguma métrica para embasar sua resposta'

# q13
class Q13(ThoughtExperiment):
    _hint = 'Use algum método de teste de normalidade. O mais simples pode ser o mais elegante!'

# q14
class Q14(ThoughtExperiment):
    _hint = 'Não vale copiar e colar da documentação, hein...'

# q15
class Q15(ThoughtExperiment):
    _hint = 'Não vale copiar e colar da documentação, hein...'

qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6,
    Q7,
    Q8,
    Q9,
    Q10,
    Q11,
    Q12,
    Q13,
    Q14,
    Q15,
], var_format='q{n}')

__all__ = list(qvars)



from learntools_dados_ufv.core.multiproblem import MultipartProblem

def backtickify(s):
    return '`{}`'.format(s)

# This is an artifact of when we had extra credit questions and wanted an option for
# a scoring mechanism that's no longer necessary
quantum_of_bonus = 1/37

def instantiate_probview(prob_cls, value_per_problem):
    # TODO: Bleh, circular import...
    from learntools_dados_ufv.core import problem_view as pv
    from learntools_dados_ufv.core.globals_binder import binder
    # NB: May eventually have some subclasses for this?
    viewer_cls = pv.ProblemView
    prob = prob_cls()
    if prob._counts_for_points:
        if prob._bonus:
            prob.point_value = quantum_of_bonus
        else:
            prob.point_value = value_per_problem
    else:
        prob.point_value = 0
    view = viewer_cls(prob, binder.readonly_globals())
    # XXX: Circular reference. :/
    # Consider using weakref (https://docs.python.org/3/library/weakref.html)
    # Also, would just have preferred a cleaner separation between these abstractions...
    prob._view = view
    return view


def bind_exercises(g, exercises, start=1, var_format='q{n}'):
    """Given the globals() dict of an exercise module, and an ordered list of
    Problem subclasses, create a sequence of variables (by default q1, q2, q3...
    but customizable via the start and var_format kwargs) referring to instantiations
    of the Problem subclasses (well, technically wrapped in ProblemView instances).
    Embed those variable assignments in the given namespace, and yield the names of
    all the new variables.

    Elements of exercises may also be None as a placeholder, in which case the 
    corresponding variable in the sequence is skipped over. e.g. [SpamProblem, None,
    EggsProblem], will generate variables q1 and q3.
    """
    denom = sum( 
            (getattr(prob, '_counts_for_points', False) and not prob._bonus) 
            for prob in exercises
            )
    try:
        value_per_problem = 1 / denom
    except ZeroDivisionError:
        value_per_problem = 1.0

    for i, prob_cls in enumerate(exercises):
        # A value of None is a placeholder. Reserve the corresponding question number, but don't create any corresponding Problem obj.
        if prob_cls is None:
            continue
        qno = i + start
        varname = var_format.format(n=qno)
        assert varname not in g
        # TODO: Probably cleaner to just pass these as sublists, rather than having
        # the MultipartProblem class start in some nascent state and get 'activated' here
        if isinstance(prob_cls, MultipartProblem):
            mpp = prob_cls
            g[varname] = mpp
            mpp._varname = varname
            for j, prob_cls in enumerate(mpp.problems):
                prob = instantiate_probview(prob_cls, value_per_problem)
                # Bleh, more properties tacked on ad-hoc outside the class.
                prob._order = '{}.{}'.format(qno, j+1)
                letter = chr(ord('a')+j)
                setattr(mpp, letter, prob)
                mpp._prob_map[letter] = prob
        else:
            pv = instantiate_probview(prob_cls, value_per_problem)
            pv._order = str(qno)
            g[varname] = pv
        yield varname
    # Bad sep of concerns, but anyways, have exercise modules also export quad alias
    # variable as alias for special Placeholder value (if it exists in their namespace
    # i.e. if they imported * from learntools_dados_ufv.core)
    if '____' in g:
        yield '____'

def format_args(fn, args):
    c = fn.__code__
    params = c.co_varnames[:c.co_argcount]
    #assert len(args) == len(params)
    return ', '.join([
        '`{}={}`'.format(param, repr(arg))
        for (param, arg) in zip(params, args)
        ])



import functools
import traceback
from collections import Counter

from IPython.display import display

from learntools_dados_ufv.core.richtext import *
from learntools_dados_ufv.core.exceptions import *
from learntools_dados_ufv.core.problem import *
from learntools_dados_ufv.core import colors, tracking

def displayer(fn):
    @functools.wraps(fn)
    def wrapped(*args, **kwargs):
        res = fn(*args, **kwargs)
        display(res)
        # Don't propagate the return to avoid double printing.
    return wrapped

def record(method):
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        self.interactions[method.__name__] += 1
        return method(self, *args, **kwargs)
    return wrapped

class ProblemView:

    _not_attempted_msg = ("Depois de atualizar o código inicial, `check()` irá"
            " dizer se seu código está correto."
    )

    def __init__(self, problem:Problem, globals_):
        self.problem = problem
        self.globals = globals_
        self.interactions = Counter()
        # The outcome of the last call to .check (as a tracking.OutcomeType).
        # Used for notebook testing.
        self._last_outcome = None

    def __getattr__(self, attr):
        """By default, expose methods of the contained Problem object if
        they're not marked private.
        """
        val = getattr(self.problem, attr)
        if not attr.endswith('_') and callable(val):
            return val
        raise AttributeError

    @property
    def questionId(self):
        # e.g. '3_MyHardProblem'
        id = self.problem.__class__.__name__
        if hasattr(self, '_order'):
            id = '{}_{}'.format(self._order, id)
        return id

    def _track_event(self, interactionType, **kwargs):
        kwargs['interactionType'] = interactionType

        if isinstance(self.problem, EqualityCheckProblem):
            kwargs['questionType'] = tracking.QuestionType.EQUALITYCHECKPROBLEM
        elif isinstance(self.problem, CodingProblem):
            kwargs['questionType'] = tracking.QuestionType.CODINGPROBLEM
        elif isinstance(self.problem, FunctionProblem):
            kwargs['questionType'] = tracking.QuestionType.FUNCTIONPROBLEM
        elif isinstance(self.problem, ThoughtExperiment):
            kwargs['questionType'] = tracking.QuestionType.THOUGHTEXPERIMENT

        problem_fields = dict(
                questionId=self.questionId,
            )
        kwargs.update(problem_fields)
        tracking.track(kwargs)

    def _track_check(self, outcome, **kwargs):
        self._last_outcome = outcome
        if outcome == tracking.OutcomeType.PASS:
            kwargs['valueTowardsCompletion'] = self.problem.point_value
        self._track_event(tracking.InteractionType.CHECK, outcomeType=outcome, **kwargs)

    @record
    @displayer
    def check(self):
        try:
            if isinstance(self.problem, CodingProblem):
                args = self._get_injected_args()
            else:
                args = ()
            self.problem.check_whether_attempted(*args)
            self.problem.check(*args)
        except NotAttempted as e:
            self._track_check(tracking.OutcomeType.UNATTEMPTED)
            return ProblemStatement(self._not_attempted_msg + ' ' + str(e))
        except (Incorrect, AssertionError) as e:
            if isinstance(e, UserlandExceptionIncorrect):
                wrapped = e.wrapped_exception
                tb_lines = traceback.format_tb(wrapped.__traceback__)
                tb_str = '\n'.join(tb_lines)
                self._track_check(tracking.OutcomeType.EXCEPTION,
                    exceptionClass=wrapped.__class__.__name__,
                    trace=tb_str,
                    failureMessage=str(e),
                    )
            else:
                self._track_check(tracking.OutcomeType.FAIL,
                    failureMessage=str(e),
                    )
            return TestFailure(str(e))
        except Uncheckable as e:
            msg = str(e) or 'Desculpa, o auto-checking não está disponível para essa questão.'
            self._track_check(tracking.OutcomeType.EXCEPTION,
                failureMessage=msg,
                exceptionClass='Uncheckable',
                trace='',
            )
            return RichText(msg, color=colors.WARN)
        else:
            self._track_check(tracking.OutcomeType.PASS)
            if hasattr(self.problem, '_congrats'):
                return Correct(self.problem._correct_message,
                               _congrats=self.problem._congrats)
            else:
                return Correct(self.problem._correct_message)

    def _get_injected_args(self):
        names = self.problem.injectable_vars
        missing = set(names) - self.globals.keys()
        if len(missing) == 0:
            return self.globals.lookup(names)
        elif len(missing) == len(names):
            # Hm, maybe RichText objects should be raisable? Or is that too much?
            raise NotAttempted("Lembre-se, você deve criar a seguinte variável {}: {}"\
                    .format('s' if len(missing) > 1 else '',
                        ', '.join('`{}`'.format(v) for v in missing)
                        )
                    )
        else:
            raise Incorrect("Você ainda precisa definir a(s) variável(is): {}".format(
                    ', '.join('`{}`'.format(v) for v in missing)
                    ))


    @record
    @displayer
    def hint(self, n=1):
        hints = self.problem.hints
        if not hints:
            msg = 'Desculpa, sem dicas para você nessa questão...'
            self._track_event(tracking.InteractionType.HINT, failureMessage=msg)
            return RichText(msg, color=colors.WARN)
        self._track_event(tracking.InteractionType.HINT)
        # TODO: maybe wrap these kinds of user errors to present them in a nicer way?
        # (e.g. LearnUserError, LearnUsageError)
        assert n <= len(hints), "Sem dicas disponíveis!"
        hint = hints[n-1]
        assert isinstance(hint, str)
        return Hint(hint, n, last=(n == len(hints)))

    @record
    @displayer
    def solution(self):
        self._track_event(tracking.InteractionType.SOLUTION)
        soln = self.problem.solution
        if isinstance(soln, RichText):
            return soln
        return Solution(soln)

    def _assert_last_outcome(self, outcome):
        self.check()
        assert self._last_outcome == outcome, ("Esperado que a última saída seja {}, mas foi {}".format(
            outcome, self._last_outcome))

    def assert_check_unattempted(self):
        self._assert_last_outcome(tracking.OutcomeType.UNATTEMPTED)

    def assert_check_failed(self):
        self._assert_last_outcome(tracking.OutcomeType.FAIL)

    def assert_check_passed(self):
        self._assert_last_outcome(tracking.OutcomeType.PASS)



import pandas as pd

from learntools_dados_ufv.core import *
from scipy.stats import percentileofscore

# q1
class Q1(CodingProblem):
    _vars = ['df', 'missing_invoice_no']
    _hints = [
        'Pesquise sobre a função `df.isna` do Pandas.',
        'Você pode usar a função `pd.isnull` com o seu dataframe, igual mostra o roteiro.',
        'A resposta é um número inteiro com a quantidade de registros.'
    ]

    _solution = ''

    def check(self, df, missing_invoice_no):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(int, missing_invoice_no, name='missing_invoice_no')
        answer = len(df[df["InvoiceNo"].isna()])
        assert_equal(answer, missing_invoice_no, name='answer')

# q2
class Q2(CodingProblem):
    _vars = ['df', 'invalid_unit_price']
    _hints = [
        'Como você pode filtrar uma coluna de um dataframe?',
        'Veja os exemplos de filtros do roteiro.',
        'A resposta é um número inteiro com a quantidade de registros.'
    ]

    _solution = ''

    def check(self, df, invalid_unit_price):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(int, invalid_unit_price, name='invalid_unit_price')
        answer = len(df[df["UnitPrice"] < 0])
        assert_equal(answer, invalid_unit_price, name='answer')

# q3
class Q3(CodingProblem):
    _vars = ['df']
    _hints = [
        'Pesquise sobre a função `df.drop` do Pandas.',
        'Execute a mesma consulta da questão anterior depois de alterar os valores para garantir que os objetos foram removidos.',
        'Restaure a versão do dataframe com `df = original_df` caso tenha alterado coisa demais.',
        'Veja os exemplos de limpeza de dados inválidos do roteiro.',
        'Você só precisa passar os índices dos valores que quer remover para a função `df.drop`.',
        'Você pode recuperar os índices de um dataframe filtrado com `df.index`.'
    ]

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')
        answer = df.drop(df.index[df['UnitPrice'] < 0])
        assert_df_equals(df, answer, name='df')

# q4
class Q4(CodingProblem):
    _vars = ['df', 'top_customer_id']
    _hints = [
        'O ID do Cliente é referente ao atributo CustomerID`',
        'Revisite o roteiro e procure por uma forma fácil de visualizar a contagem de valores mais frequentes.',
        'A resposta é um inteiro com o valor do atributo CustomerID.'
    ]

    _solution = ''

    def check(self, df, top_customer_id):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(int, top_customer_id, name='top_customer_id')
        answer = 17841
        assert_equal(top_customer_id, answer, name='top_customer_id')

# q5
class Q5(CodingProblem):
    _vars = ['df', 'most_expensive_product']
    _hints = [
        'Tente recuperar primeiro qual o maior valor da coluna de Preço Unitário.',
        'Você pode seguir o mesmo caminho da questão anterior, mas existem formas mais simples e elegante usando um simples filtro.',
        'Talvez após filtrar você precise usar a função `df.iloc` para recuperar a primeira linha do registro, e acessar a coluna desejada depois.',
        'A resposta é uma string com o valor da descrição do produto.'
    ]

    _solution = ''

    def check(self, df, most_expensive_product):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(str, most_expensive_product, name='most_expensive_product')
        answer = 'Manual'
        assert_equal(most_expensive_product, answer, name='most_expensive_product')

# q6
class Q6(CodingProblem):
    _vars = ['df', 'germany_orders_percentage']
    _hints = [
        'Utilize os filtros como detalhado no roteiro para remover o país `United Kingdom` da seleção.',
        'Veja a função `df.value_counts` do Pandas.',
        'Você pode chamar a função `value_counts` diretamente em uma coluna (ex: `df["Coluna"].value_counts()).',
        'Você pode passar o parâmetro `normalize=True` na função `value_counts`, como mostra o roteiro.',
        'O retorno da função `value_counts` é uma Serie pandas. Basta acessá-lo passando um índice como `serie["MeuIndice"]`.',
        'A resposta é um float representando a porcentagem (lembre-se de multiplicar por 100!).'
    ]

    _solution = ''

    def check(self, df, germany_orders_percentage):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(float, germany_orders_percentage, name='germany_orders_percentage')
        answer = df[df['Country'] != 'United Kingdom'].Country.value_counts(normalize=True)['Germany'] * 100
        assert_equal(germany_orders_percentage, answer, name='germany_orders_percentage')

# q7
class Q7(CodingProblem):
    _vars = ['df', 'most_items_order']
    _hints = [
        'Recupere primeiro qual instância tem a maior quantidade de produtos (coluna Quantity).',
        'Utilize filtros para acessar o dataframe que tem a coluna `Quantity` com o valor recuperado da dica anteiror.',
        'A resposta é um dataframe filtrado.'
    ]

    _solution = ''

    def check(self, df, most_items_order):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(pd.DataFrame, most_items_order, name='most_items_order')
        answer = df[df['Quantity'] == df['Quantity'].max()]
        assert_df_equals(most_items_order, answer, name='most_items_order')

# q8
class Q8(CodingProblem):
    _vars = ['df', 'most_expensive_order']
    _hints = [
        'Pense numa compra que você faz em algum comércio. Como é calculado o valor final?.',
        'Veja no roteiro como criar uma coluna composta, pode ser útil.',
        'Pense que cada instância do dataframe é um registro de um produto em uma compra.',
        'O que identifica uma compra no dataframe?',
        'Tente agrupar os registros pelo identificador da compra.',
        'Com os registros agrupados, veja sobre a função `dfgroupedby.sum` do Pandas',
        'Tente somar o valor pago por todos os produtos em uma compra.',
        'Tente ordenar as instâncias da compra pelo seu valor pago e veja qual o valor da coluna InvoiceNo',
        'A resposta é um inteiro com o valor do identificador da compra (InvoiceNo).'
    ]

    _solution = ''

    def check(self, df, most_expensive_order):
        assert_isinstance(pd.DataFrame, df, name='df')
        assert_isinstance(int, most_expensive_order, name='most_expensive_order')
        answer = 581483
        assert_equal(most_expensive_order, answer, name='most_expensive_order')

qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
    Q6,
    Q7,
    Q8
], var_format='q{n}')

__all__ = list(qvars)



from abc import ABC, abstractmethod
from typing import List
import functools

from learntools_dados_ufv.core.richtext import *
from learntools_dados_ufv.core.exceptions import NotAttempted, Uncheckable, UserlandExceptionIncorrect
from learntools_dados_ufv.core import utils, asserts, constants

# I'm sure there's a more elegant way to do this.
# Some kind of decorator on top of property?
def optionally_plural_property(obj, name):
    single_attr = getattr(obj, name, None)
    plural_attr = getattr(obj, name + 's', None)
    assert single_attr is None or plural_attr is None, ("Subclass should not implement"
            " both {0} and {0}s").format(name)
    if single_attr is not None:
        return [single_attr]
    elif plural_attr is not None:
        return plural_attr
    else:
        return []

class Problem(ABC):

    _solution = ''

    # used for valueTowardsCompletion bookkeeping
    _counts_for_points = True
    _bonus = False

    @property
    def solution(self):
        return self._solution

    @property
    def hints(self) -> List[str]:
        return optionally_plural_property(self, '_hint')

    @property
    def _correct_message(self):
        if (
                self.show_solution_on_correct 
                or (self.show_solution_on_correct is None 
                    and isinstance(self.solution, str)
                    )
                ):
            return '\n\n' + self.solution
        else:
            return ''

    @abstractmethod
    def check(self, *args):
        """If this method runs without exceptions, it indicates that checking passed
        and the solution is correct. To indicate other outcomes, implementations of 
        this method should raise one of the following:
        - Uncheckable: If this problem explicitly has no checking logic.
        - NotAttempted: If it seems the problem hasn't been attempted (i.e. the 
            starter code hasn't been modified.
        - Incorrect, AssertionError: If there's a problem with the user's solution.

        Any messages attached to these exceptions will be passed on in the message shown
        to the user.
        """
        pass

    def check_whether_attempted(self, *args):
        pass

class ThoughtExperiment(Problem):

    show_solution_on_correct = True
    def check(self, *args):
        pass

# TODO: apply directly to EqualityCheckProblem.check etc.?
def injected(method):
    """A decorator for (custom) methods of Problem subclasses which want to receive
    injected values from the student's notebook as arguments - in the same way that
    .check(), .check_whether_attempted() etc. are automatically supplied injected
    args in CodingProblem subclasses.

    Injected methods may also receive additional, explicit, user-supplied arguments.
    They should come after any injected args.
    """
    @functools.wraps(method)
    def wrapped(self, *args, **kwargs):
        # More muddying of the waters btwn Problem and ProblemView :/
        # XXX: Handling of unset variables. This call may throw NotAttempted/Incorrect.
        injargs = self._view._get_injected_args()
        # Sometimes there are user-supplied args to pass on in addition to the
        # ones we're injecting (see python.ex3 Blackjack problem for an example of this)
        # Injected args first, then any additional ones.
        newargs = list(injargs) + list(args)
        return method(self, *newargs, **kwargs)

    return wrapped

class CodingProblem(Problem):
    # What do we show when the user calls .check() and their code is correct?
    # False: just tell them they were right
    # True: tell them they were right, and show them the solution (as if they had 
    #       called problem.solution())
    # None: Use the default heuristic, which shows the solution iff solution is
    #       a string (i.e. not an instance of CodeSolution).
    show_solution_on_correct = None

    _var = None
    _vars = None
    # Can optionally set _default_values class attr (for purposes of NotAttempted checking)
    
    @property
    def injectable_vars(self) -> List[str]:
        return optionally_plural_property(self, '_var')

    def check_whether_attempted(self, *args):
        # TODO: copy-paste from EqualityCheckProblem.check_whether_attempted
        varnames = self.injectable_vars
        def _raise_not_attempted():
            raise NotAttempted("You need to update the code that creates"
                    " variable{} {}".format('s' if len(varnames) > 1 else '',
                        ', '.join(map(utils.backtickify, varnames))))
        # First, check whether any vars have placeholder values
        for (var, val) in zip(varnames, args):
            # NB: Yoda condition to ensure that it's PlaceholderValue's __eq__ that gets called.
            if constants.PLACEHOLDER == val:
                _raise_not_attempted()


class EqualityCheckProblem(CodingProblem):
    """A problem which is considered solved iff some user-defined variables 
    are equal to some groundtruth expected values.

    The conventional way for subclasses to specify expected values is with a _expected
    member, containing a list of expected values (of the same length as _vars and in
    the same order).

    In the common case where there is only one variable of interest (_var), subclasses
    can set _expected to be a simple scalar (rather than wrapping the value in a list of
    length 1). (Special case: to avoid ambiguity, if the expected value is itself a list 
    of length 1, it must be wrapped)
    """

    @property
    def expected(self):
        """A list of expected values. Matches length and order of _var/_vars
        """
        ex = self._expected
        if len(self.injectable_vars) == 1:
            # Don't wrap length-1 lists (i.e. assume that ex[0] is the expected value
            # of our single variable of interest, rather than ex itself)
            if (isinstance(ex, list) or isinstance(ex, tuple)) and len(ex) == 1:
                return ex
            else:
                return [ex]
        else:
            assert len(ex) == len(self.injectable_vars)
            return ex

    def check(self, *args):
        for (var, actual, expected) in zip(self.injectable_vars, args, self.expected):
            asserts.assert_equal(actual, expected, var=var,
                    failure_factory=getattr(self, '_failure_message', None)
                    )

    def check_whether_attempted(self, *args):
        varnames = self.injectable_vars
        def _raise_not_attempted():
            raise NotAttempted("You need to update the code that creates"
                    " variable{} {}".format('s' if len(varnames) > 1 else '',
                        ', '.join(map(utils.backtickify, varnames))))
        # First, check whether any vars have placeholder values
        for (var, val) in zip(varnames, args):
            # NB: Yoda condition to ensure that it's PlaceholderValue's __eq__ that gets called.
            if constants.PLACEHOLDER == val:
                _raise_not_attempted()
        if not hasattr(self, '_default_values'):
            return
        for var, val, default in zip(
                varnames, args, self._default_values
                ):
            try:
                neq = val != default
            except:
                # If we get an exception comparing the actual value to the expected,
                # we can reasonably infer they're not equal.
                neq = True
            # If the actual value is an ndarray or DataFrame or something, the result of != may
            # be something more complicated than a bool. If this is the case, let's assume the
            # actual value differs from the default. (Default values will probably never be anything
            # as complicated an ndarray/df/series anyways)
            if not isinstance(neq, bool):
                return
            if neq:
                return
        # It'd be kind of odd if a EqualityCheckProblem didn't have any associated
        # vars, but I guess it's not worth raising a fuss over...
        if len(args):
            _raise_not_attempted()


class FunctionProblem(CodingProblem):

    # Must have a single _var, corresponding to a function name
    
    # List of (input, expected_output) pairs, where input may be a scalar or tuple of args.
    _test_cases = []
    
    @classmethod
    def check_whether_attempted(cls, fn):
        # Not sure if inspect.getsource() will work reliably? Seems like it does
        # work okay with ipython notebooks. 
        def dummy(): 
            pass
        def dummy_w_docstring():
            """blah blah fishcakes"""
            pass
        # NB: looks like in terms of bytecode, a "pass" is treated like "return None", meaning this will 
        # possibly match other fns that just return some constant. Could be a little dangerous/brittle.
        src = lambda f: f.__code__.co_code
        if src(fn) in (src(dummy), src(dummy_w_docstring)):
            raise NotAttempted

    def check(self, fn):
        assert self._test_cases, "Oops, someone forgot to write test cases."
        for args, expected in self._test_cases:
            orig_args = args
            # Wrap in tuple if necessary
            if not isinstance(args, tuple):
                args = args,
            # XXX: ugh, need to guard against mutation :(
            # Beware of more exotic mutable types (which lack a copy method, or which
            # require a deep copy). Maybe this shouldn't be handled at this level -
            # maybe cleaner to have a method that can repeatedly spit out fresh lists
            # of test cases every time it's called, rather than having test cases be
            # a static attribute.
            args = [(arg.copy() if hasattr(arg, 'copy') else arg) for arg in args]
            orig_args = [(arg.copy() if hasattr(arg, 'copy') else arg) for arg in args]
            try:
                actual = fn(*args)
            except Exception as e:
                raise UserlandExceptionIncorrect(e, orig_args)
            assert not (actual is None and expected is not None), ("Got a return value of `None`"
                    " given {}, but expected a value of type `{}`. (Did you forget a `return` statement?)"
                    ).format(utils.format_args(fn, orig_args), type(expected).__name__)
            assert actual == expected, ("Expected return value of `{}` given {},"
                    " but got `{}` instead.").format(
                            repr(expected), utils.format_args(fn, orig_args), repr(actual))


__all__ = ['Problem', 'EqualityCheckProblem', 'FunctionProblem',
        'ThoughtExperiment', 'CodingProblem',
        ]



from learntools_dados_ufv.core import *

# q1
class Q1(ThoughtExperiment):
    _hints = [
        'Revisite a lista de análise exploratória. Quais informações você consegue extrair dessa base?',
        'Tente visualizar um conjunto de gráficos par-a-par de cada coluna. No roteiro tem um exemplo.',
        'Tente traçar uma linha no gráfico par a par. É fácil de ter uma reta que intercepta a maioria dos pontos?',
    ]

# q2
class Q2(ThoughtExperiment):
    _hints = [
        'Identifique e atribua os valores de X e Y para ser alimentado em uma regressão linear.',
        'Talvez você precise fazer um `reshape` com o eixo X, assim como está no roteiro.',
        'Utilize do `LinearRegression` do sklearn como apresentado no roteiro.',
        'Pesquise sobre o `coef_` e `score` do LinearRegression',
    ]

# q3
class Q3(ThoughtExperiment):
    _hints = [
        'Identifique e atribua os valores de X e Y para ser alimentado em uma regressão linear.',
        'Talvez você precise fazer um `reshape` com o eixo X, assim como está no roteiro.',
        'Utilize do `LinearRegression` do sklearn como apresentado no roteiro.',
        'Pesquise sobre o `coef_` e `score` do LinearRegression',
    ]

# q4
class Q4(ThoughtExperiment):
    _hints = [
        'Identifique e atribua os valores de X e Y para ser alimentado em uma regressão linear.',
        'Talvez você precise fazer um `reshape` com o eixo X, assim como está no roteiro.',
        'Lembre-se que você precisa passar múltiplas colunas como parte do eixo X',
        'Utilize do `LinearRegression` do sklearn como apresentado no roteiro.',
        'Pesquise sobre o `coef_` e `score` do LinearRegression',
    ]

# q5
class Q5(ThoughtExperiment):
    _hints = [
        'Tente fazer uma regressão linear com todas as colunas.',
        'Interprete os scores e coeficientes retornados em uma regressão linear.',
        'O que é mais importante para ter uma alta correlação?',
    ]


qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
    Q4,
    Q5,
], var_format='q{n}')

__all__ = list(qvars)



class MultipartProblem:
    """A container for multiple related Problems grouped together in one 
    question. If q1 is a MPP, its subquestions are accessed as q1.a, q1.b, etc.
    """
    
    def __init__(self, *probs):
        self.problems = probs
        # Mapping from string identifier (e.g. 'a', 'b'...) to ProblemView
        self._prob_map = {}

    def _repr_markdown_(self):
        return repr(self)

    def __repr__(self):
        varname = self._varname
        part_names = ['`{}.{}`'.format(varname, letter) for letter in self._prob_map]
        return """Essa questão possui {} partes. Essas partes podem ser acessadas 
como {}. Por exemplo, para pegar uma dica na parte a, digite `{}.a.hint()`.`""".format(
            len(self._prob_map), ', '.join(part_names), varname
        )




import logging

# This method captures a reference to the globals dict so it can be seen by checking code

class Binder:
    
    def __init__(self):
        self.bound = False
        self.g = None

    def bind(self, global_vars):
        if self.bound:
            if id(global_vars) == id(self.g):
                logging.warn("Ignoring repeated attempt to bind to globals")
            else:
                raise Exception("Attempted to bind different vars to already-bound binder")
        else:
            self.g = global_vars
            self.bound = True

    def readonly_globals(self):
        return ReadOnlyGlobals(self.g)

class ReadOnlyGlobals:
    """Readonly wrapper around captured globals dict, plus some convenience methods.
    """

    def __init__(self, g):
        self.g = g

    def __getitem__(self, key):
        # TODO: maybe wrap KeyError in some custom exception? 
        # Or maybe there should be two ways of looking up a key, one of which has
        # the semantics "if this key is missing, the user did something terribly wrong"
        return self.g[key]

    def __contains__(self, key):
        return key in self.g

    def keys(self):
        return self.g.keys()

    def lookup(self, keys):
        return [self.g[k] for k in keys]

    # No __setitem__ 

binder = Binder()



"""
Some assertion helpers used in Problem.check implementations.

These are most heavily used in EqualityCheckProblems. They have nice error messages, and thus could be
used in how we code check methods in CodingProblems, though that wasn't a widespread practice when
this comment was written in Dec 2019.
"""

import os
import numbers
import math
import functools
import logging

import pandas as pd
import numpy as np

def name_or_var(assert_fn):
    """
    Assert helpers marked with this decorator take a "name" argument, which is a
    markdown string describing the actual value being checked. It may be a generic
    name like "dataframe", or a prose description like "the result of calling `circle_area`",
    or a variable name wrapped in backticks.

    The latter case is extremely common, so this decorator adds an optional "var" keyword-arg
    to the function. Passing var="foo" is a convenient shorthand for name="`foo`". (If you
    pass a value for "var", you should not also pass a value for "name".)

    Example:

    @name_or_var
    def assert_negative(actual, name):
        assert actual < 0, "{} should have been negative, but was actually {}".format(name, actual)

    # The following are all valid calls
    assert_negative(x, "Bank balance")
    assert_negative(x, name="`Bank.balance` attribute")
    assert_negative(x, var="bank_balance") # Equivalent to assert_negative(x, "`bank_balance`")
    """
    @functools.wraps(assert_fn)
    def wrapped(*args, **kwargs):
        var = kwargs.pop('var', None)
        if var:
            if 'name' in kwargs:
                logging.warn("Function {} received values for keyword args name *and* var. Overwriting original name kwarg.".format(
                    assert_fn.__name__))
            kwargs['name'] = '`{}`'.format(var)
        return assert_fn(*args, **kwargs)
    return wrapped

@name_or_var
def assert_equal(actual, expected, name, failure_factory=None):
    """Assert a protean notion of equality specific to the use case of learntools_dados_ufv
    checking. Subclasses of EqualityCheckProblem ultimately use this function
    in their check method.

    Includes special cases for several types of expected values, including Pandas
    objects, ndarrays, and floats.
    """
    # We default to == comparison, but have special cases for certain data types.
    if isinstance(expected, float):
        assert isinstance(actual, numbers.Number), \
            "Esperado que {} seja um número, mas possui valor `{!r}` (type = `{}`)".format(
                name, actual, type(actual).__name__)
        check = math.isclose(actual, expected, rel_tol=1e-06)
    elif isinstance(expected, pd.DataFrame):
        assert_df_equals(actual, expected, name)
        return
    elif isinstance(expected, pd.Series):
        assert_series_equals(actual, expected, name)
        return
    elif isinstance(actual, np.ndarray) or isinstance(expected, np.ndarray):
        check = np.array_equal(actual, expected)
    else:
        check = actual == expected
    if failure_factory:
        # This optional kwarg lets the caller pass a function to generate a custom
        # failure message. Currently only used in the Python ex1 favourite color question.
        _failure_message = failure_factory(name, actual, expected)
    else:
        _failure_message = "Valor incorreto para {}: `{}`".format(
                name, repr(actual))
    assert check, _failure_message

@name_or_var
def assert_has_columns(df, cols, name="dataframe", strict=False):
    """Assert that the given dataframe contains columns with the given names.
    If strict is True, then assert it has *only* those columns.
    """
    for col in cols:
        assert col in df.columns, "Esperado que {} tenha a coluna `{}`".format(
                name, col
                )
    if strict:
        for col in df.columns:
            msg = "Coluna inesperada em {}: `{}`".format(name, col)
            assert col in cols, msg

@name_or_var
def assert_isinstance(cls, actual, name):
    assert isinstance(actual, cls), "Esperado que a variável `{}` tenha o tipo `{!r}` mas tem o tipo `{!r}`".format(name, cls, type(actual))

@name_or_var
def assert_is_one_of(actual, options, name):
    msg = "Valor incorreto para {}: `{!r}`".format(name, actual)
    assert actual in options, msg

@name_or_var
def assert_len(thing, exp_len, name):
    """Assert that the given thing has the given length.

    PRECONDITION: the thing implements __len__
    """
    actual = len(thing)
    assert actual == exp_len, "Esperado que {} tenha tamanho {}, mas possui tamanho {}".format(
            name, exp_len, actual,
            )

def assert_file_exists(path):
    if '/' in path:
        pp = 'at path'
    else:
        pp = 'with name'
    msg = "Expected file to exist {} `{}`".format(pp, path)
    assert os.path.exists(path), msg
    assert os.path.isfile(path), "Expected `{}` to be a file".format(path)

@name_or_var
def assert_df_equals(actual, exp, name="dataframe"):
    assert_isinstance(pd.DataFrame, actual, name)
    assert len(actual) == len(exp), "Esperado que {} tenha tamanho {} mas possui tamanho {}".format(
        name, len(exp), len(actual))
    # Only verify that the first n records match - I guess this could be slow if 
    # our dataframes have hundreds of thousands of rows. This *could* bite us, though
    # it seems unlikely to cause a false negative.
    lim = 100
    actual_sub = actual.head(lim)
    exp_sub = exp.head(lim)
    eq = actual_sub.equals(exp_sub)
    if eq:
        return
    # Okay, so they weren't equal. Let's try to come up with a helpful message about
    # how they disagree. (TODO: I wonder if someone has written code for this already
    # somewhere. Seems like it would be useful for unit testing, for example.)
    assert_has_columns(actual, exp.columns, name, strict=True)
    # TODO: Check index equality? Check equality of values column by column? 
    # Check dtype match per column. (This is something that df.equals cares about,
    # though in some cases that might be stricter than what we want. e.g. we might 
    # not care if a column has values [1, 2, 3] in expected and [1., 2., 3.] in actual)
    assert False, "Valor incorreto para {}".format(name)

@name_or_var
def assert_series_equals(actual, exp, name="series"):
    assert_isinstance(pd.Series, actual, name=name)
    assert len(actual) == len(exp), "Esperado que {} tenha tamanho {} mas possui tamanho {}".format(
        name, len(exp), len(actual))
    lim = 100
    actual_sub = actual.head(lim)
    exp_sub = exp.head(lim)
    eq = actual_sub.equals(exp_sub)
    if eq:
        return
    # TODO: More checks
    assert False, "Valor incorreto para {}".format(name)

# For star import purposes, export only names that begin with assert (i.e. our helper fns)
__all__ = [name for name in dir() if name.startswith('assert')]



from . import colors

def colorify(text, color):
    return '<span style="color:{}">{}</span>'.format(color, text)

class RichText:
    """Crucially this defines methods for both rich and plaintext representations.
    If displayed from a code cell, it will render the nice rich output. If in the console,
    we'll fall back to the simple representation.
    """
    def __init__(self, txt, color=None, **kwargs):
        self.txt = txt
        self.color = color
        self.kwargs = kwargs

    def _repr_markdown_(self):
        if self.color:
            return colorify(self.txt, self.color)
        return self.txt

    def __repr__(self):
        return self.txt

class PrefixedRichText(RichText):
    """A RichText message prefixed with some label (which may optionally be styled
    with a distinct color).
    """
    label_color = 'black' # This can be overridden with any valid CSS color
    @property
    def label(self):
        # If no _label attribute is present, fall back to the class name.
        if hasattr(self, '_label'):
            return self._label
        return self.__class__.__name__

    def _repr_markdown_(self):
        if not self.txt:
           return colorify(self.label, self.label_color)
        pre = colorify(self.label+':', self.label_color)
        return pre + ' ' + self.txt

    def __repr__(self):
        if not self.txt:
            return self.label
        return self.label + ':' + ' ' + self.txt


class Hint(PrefixedRichText):
    label_color = colors.HINT
    def __init__(self, txt, n=1, last=True):
        self.n = n
        # Is this one of a series of hints?
        self.is_multi = n > 1 or not last
        if not last:
            # TODO: Would be nice to be able to reference the actual name
            # associated with the ProblemView, e.g. `q2.hint(2)`
            coda = '\n(Para outra dica, chame `.hint({})`)'.format(n+1)
            txt += coda
        super().__init__(txt)

    @property
    def label(self):
        if self.is_multi:
            return 'Dica {}'.format(self.n)
        return 'Dica'

class Correct(PrefixedRichText):
    @property
    def _label(self):
        return self.kwargs.get('_congrats', 'CORRETO')
    label_color = colors.CORRECT

class Solution(PrefixedRichText):
    label_color = colors.SOLUTION

class CodeSolution(Solution):
    """A solution consisting entirely of Python code. We wrap this in a
    syntax-highlighted code block.
    """
    _label = 'Solução'

    def __init__(self, *lines):
        """As a convenience, may pass in one string per line of code, rather than
        one big multi-line string.
        """
        txt = '\n'.join(lines)
        wrapped = "\n```python\n{}\n```".format(txt)
        super().__init__(wrapped)

    @classmethod
    def load(cls, path):
        """Return a CodeSolution containing the code located in the source file
        at the given path.
        """
        with open(path) as f:
            lines = f.readlines()
            # Strip trailing newlines (cause constructor adds them back...)
            lines = [line[:-1] for line in lines
                    # Hack
                    if not line.startswith('from learntools_dados_ufv.python.solns')
                    ]
            return cls(*lines)

class TestFailure(PrefixedRichText):
    label_color = colors.INCORRECT
    _label = 'INCORRETO'

class ProblemStatement(PrefixedRichText):
    label_color = colors.INFO
    _label = 'VERIFICAÇÃO'



import learntools_dados_ufv
from setuptools import setup
from setuptools import find_packages

setup(name='learntools_dados_ufv',
      version=learntools_dados_ufv.__version__,
      description='Pacote auxiliar para correção e validação de práticas em dados',
      url='http://github.com/gfviegas/learntools-dados-ufv',
      author='Gustavo Viegas, Dan Becker',
      author_email='gustavo.viegas@ufv.br',
      include_package_data=True,
      license='Apache 2.0',
      packages=find_packages(),
      zip_safe=True)



import enum
from IPython.display import display, Javascript
import json

import learntools_dados_ufv

# If set to True, then echo logged events as output.
DEBUG = False

class InteractionType(enum.Enum):
    CHECK = 1
    HINT = 2
    SOLUTION = 3

class OutcomeType(enum.Enum):
    PASS = 1
    FAIL = 2
    EXCEPTION = 3
    UNATTEMPTED = 4

class QuestionType(enum.Enum):
    EQUALITYCHECKPROBLEM = 1
    CODINGPROBLEM = 2
    FUNCTIONPROBLEM = 3
    THOUGHTEXPERIMENT = 4

_EVENT_DEFAULTS = dict(
        learnToolsVersion = str(learntools_dados_ufv.__version__),
        valueTowardsCompletion = 0.0,
        failureMessage = '',
        exceptionClass = '',
        trace = '',
)

def track(event):
    # TODO: could be nice to put some validation logic here.
    for k, v in _EVENT_DEFAULTS.items():
        event.setdefault(k, v)

    # Convert enum values to plain ints
    interaction_type = event['interactionType']
    assert interaction_type in InteractionType
    event['interactionType'] = interaction_type.value
    outcome_type = event.get('outcomeType', None)
    if outcome_type:
        assert outcome_type in OutcomeType
        event['outcomeType'] = outcome_type.value
    else:
        assert interaction_type != InteractionType.CHECK, "Check events must have an OutcomeType set: {!r}".format(event)
        # Looks like we need to set some dummy value here (even if this field isn't applicable because this
        # isn't a check event. Setting outcomeType to None/null resulted in 500 errors.)
        event['outcomeType'] = 4

    question_type = event.get('questionType', None)
    if question_type:
        assert question_type in QuestionType
        event['questionType'] = question_type.value

    message = dict(jupyterEvent='custom.exercise_interaction',
            data=event)
    js = 'parent.postMessage({}, "*")'.format(json.dumps(message))
    display(Javascript(js))
    if DEBUG:
        debug_js = 'console.log({})'.format(json.dumps(message))
        display(Javascript(debug_js))
        display(message)




from urllib import response
import pandas as pd
import numpy as np


from mlxtend.preprocessing import TransactionEncoder
from mlxtend.frequent_patterns import apriori
from mlxtend.frequent_patterns import association_rules

from learntools_dados_ufv.core import *

root_df = pd.read_csv('./Market_Basket_Optimisation.csv',header=None)

def run_df_processing(steps=0):
    answer = root_df.copy()
    if (steps == 0): 
        return answer
    
    transactions_array = []
    for i in range(len(answer)):
        transactions_array.append([str(answer.values[i,j]) for j in range(len(answer.columns))])
        
    transactions_array = np.array(transactions_array)
    transactions_oht = TransactionEncoder()
    transactions = transactions_oht.fit(transactions_array).transform(transactions_array)
    
    answer = pd.DataFrame(transactions, columns=transactions_oht.columns_)
    answer.drop(['nan'],axis=1,inplace=True)

    if (steps == 1): 
        return answer
    
    answer = apriori(answer, min_support=0.01, use_colnames=True)

    if (steps == 2): 
        return answer
    
    answer = association_rules(answer, metric='confidence', min_threshold=0.3)
    
    if (steps == 3): 
        return answer
    
    if (steps == 4):
        return answer[(answer['support'] >= 0.015) & (answer['confidence'] >= 0.4)]
    
    if (steps == 5):
        return answer[(answer['lift'] >= 2) & (answer['confidence'] >= 0.5)]
    

# q1
class Q1(CodingProblem):
    _var = 'df'
    #_hint = 'Utilize a função `pd.read_csv` com o caminho do arquivos `pesos_populacao.csv`. Note que não há cabeçalhos no arquivo.'

    _solution = ''

    def check(self, df):
        assert_isinstance(pd.DataFrame, df, name='df')
        answer = run_df_processing(1)
        assert_df_equals(df, answer, name='df')


# q2.a
class Q2A(CodingProblem):
    _var = 'frequent_itemsets'
    _hints = [
        'Utilize a função `apriori` como mostrado no roteiro.',
        'Você deve passar os paramêtros na descrição do problema.',
        'Você adicionou os parâmetros `use_colnames` e `min_support`?',
    ]

    _solution = ''

    def check(self, frequent_itemsets):
        assert_isinstance(pd.DataFrame, frequent_itemsets, name='frequent_itemsets')
        answer = run_df_processing(2)
        assert_df_equals(frequent_itemsets, answer, name='frequent_itemsets')

# q2.b
class Q2B(CodingProblem):
    _var = 'rules'
    _hints = [
        'Utilize a função `association_rules` como mostrado no roteiro.',
        'Você deve passar os paramêtros na descrição do problema.',
        'Você adicionou o parâmetro `min_threshold`?',
        'A metric passada como parâmetro é a `confidence`?',
    ]

    _solution = ''

    def check(self, rules):
        assert_isinstance(pd.DataFrame, rules, name='rules')
        answer = run_df_processing(3)
        assert_df_equals(rules, answer, name='rules')

Q2 = MultipartProblem(Q2A, Q2B)

# q3.a
class Q3A(CodingProblem):
    _var = 'rules_a'
    _hints = [
        'Tente fazer primeiro o filtro de suporte.',
        'Monte agora o filtro de confidence',
        'Junte os dois filtros utilizando o operador and (&)'
        
    ]

    _solution = ''

    def check(self, rules_a):
        answer = run_df_processing(4)
        assert_equal(rules_a, answer, name='rules_a')

# q3.b
class Q3B(CodingProblem):
    _var = 'rules_b'
    _hints = [
        'Tente fazer primeiro o filtro do lift.',
        'Monte agora o filtro de confidence',
        'Junte os dois filtros utilizando o operador and (&)'
        
    ]

    _solution = ''

    def check(self, rules_b):
        answer = run_df_processing(5)
        assert_equal(rules_b, answer, name='rules_b')


Q3 = MultipartProblem(Q3A, Q3B)


qvars = bind_exercises(globals(), [
    Q1,
    Q2,
    Q3,
  
], var_format='q{n}')

__all__ = list(qvars)



class NotAttempted(Exception):
    pass

class Incorrect(Exception):
    pass

class UserlandExceptionIncorrect(Incorrect):
    """Raised when a user's solution is incorrect because it raised an (unexpected)
    exception when called.
    """
    def __init__(self, exception, args):
        self.wrapped_exception = exception
        self.msg  = ("When calling your function with arguments `{!r}`, Python"
                " raised the following exception... **`{}: {}`**").format(
                        args, exception.__class__.__name__, exception)
    def __str__(self):
        return self.msg

class Uncheckable(Exception):
    pass
